/*****************************************************************************
 * grab-avi.c: avi interface, reading frames from an avi file, only for test
 *  THE AVI FILE MUST BE IN XVID FORMAT.
 *****************************************************************************
 * $Id: grab-avi.c,v 1.7 2003/09/17 17:53:44 alainjj Exp $
 *****************************************************************************
 * Copyright (C) 2003 Alainjj
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/times.h>

#include "config.h"
#ifdef HAVE_XVID

#include <xvid.h>
#include "strtab.h"
#include "grab.h"
#include "avilib.h"
#include "memcpy.h"

void sigint_handler(int sig);

extern void video_set_capture_size(int w, int h);
extern int nbufs;

static int grab_open (struct device_t *device, int sw, int sh,
                      int format, int pixmap, void *base, int width);
static int grab_close ();
static int grab_overlay (int x, int y, int width, int height, int format,
                         struct OVERLAY_CLIP *oc, int count);
static void *grab_scr (void *dest, int width, int height, int single,
                       int dest_xsize);
static void *grab_one (int format, int width, int height);
static int grab_tune (unsigned long freq);
static int grab_tuned ();
static int grab_input (int input, int norm);
static int grab_picture (int color, int bright, int hue, int contrast);
static int grab_audio (int mute, int volume, int *mode);
static int is525(int norm);
static int issecam(int norm);
static void* get_buf(int i);

/* variable specific for grab-avi */
char *avifilename;
int avibench = 0;
static avi_t *   avifile = NULL;
static XVID_DEC_PARAM xvid_dec_p;
int avi_out_bytes,avi_out_format;
void **buffers;
extern int img;
static char* bufcompress;
static unsigned char *rgb16tohi240 = NULL;
static double delay; // delay between two images
static clock_t t_start;
struct tms start;


static struct STRTAB norms[] = {
  {0,   "PAL"},
  {1,  "NTSC"},
  {2, "SECAM"},
  {-1, NULL}
};

static struct STRTAB inputs[] = {
  {0,   "Television"},
  {-1, NULL}
};

struct GRABBER grab_avi = {
  "avi",
  VIDEO_RGB16 | VIDEO_RGB24 | VIDEO_RGB32,
  0,
  norms, inputs,
  grab_open,
  grab_close,
  grab_overlay,
  grab_scr,
  grab_one,
  grab_tune,
  grab_tuned,
  grab_input,
  grab_picture,
  grab_audio,
  is525,
  issecam,
  get_buf
};

extern unsigned short hi240_to_rgb[256][3];
static void build_rgb16tohi240(void) {
  int i,j,j2=0;
  rgb16tohi240 = malloc(65536*sizeof(unsigned char));
  for(i=0;i<65536;i++) {
    int b=i&0x1F,g=(i&0x7E0)>>5,r=(i&0xF800)>>11;
    int min=1<<30,diff;
    b=65535.0/31*b;g=65535.0/63*g;r=65535.0/31*r;
    for(j=16;j<=240;j++) {
      diff=299*abs(r-hi240_to_rgb[j][0])+
	567*abs(g-hi240_to_rgb[j][1])+
	114*abs(b-hi240_to_rgb[j][2]);
      if(diff<min) {j2=j;min=diff;}
    }
    rgb16tohi240[i]=j2;
  }
}


static int
grab_open(struct device_t *device, int sw, int sh,
          int format, int pixmap, void *base, int bpl)
{
  int i;
  int avi_out_alloc_bytes;
  XVID_INIT_PARAM ini;
  avifile = AVI_open_input_file(avifilename, 1);
  if(avifile == NULL) {
    AVI_print_error("AVI_open_input_filename");
    return -1;
  }
  if(debug)
    fprintf(stderr,
	    "AVI: width=%ld height=%ld\n",avifile->width,avifile->height);
  video_set_capture_size(avifile->width, avifile->height);
  memset(&ini,0,sizeof(XVID_INIT_PARAM));
  memset(&xvid_dec_p,0,sizeof(XVID_DEC_PARAM));

  if(xvid_init(NULL, 0, &ini, NULL)) {
    fprintf(stderr,"initialization xvid failed\n");
    return -1;
  }
  if(ini.api_version != API_VERSION) {
    fprintf(stderr,"xvid has changed since compilation, recompile !\n");
    return -1;
  }

  xvid_dec_p.width = avifile->width;
  xvid_dec_p.height = avifile->height;

  if(xvid_decore(NULL, XVID_DEC_CREATE, &xvid_dec_p, NULL)) {
    fprintf(stderr, "xvid init failed\n");
    return -1;
  }
  
  switch (pixmap)
    {
    case VIDEO_RGB08:
      avi_out_format = XVID_CSP_RGB565;
      avi_out_bytes = 1;
      build_rgb16tohi240();
      break;
    case VIDEO_GRAY:
      avi_out_format = XVID_CSP_YV12;
      avi_out_bytes = 1;
      break;
    case VIDEO_RGB15:
      avi_out_format  = XVID_CSP_RGB555;
      avi_out_bytes = 2;
      break;
    case VIDEO_RGB16:
      avi_out_format = XVID_CSP_RGB565;
      avi_out_bytes = 2;
      break;
    case VIDEO_RGB24:
      avi_out_format = XVID_CSP_RGB24;
      avi_out_bytes = 3;
      break;
    case VIDEO_RGB32:
      avi_out_format = XVID_CSP_RGB32;
      avi_out_bytes = 4;
      break;
    case VIDEO_YUY2:
      avi_out_format = XVID_CSP_YUY2;
      avi_out_bytes = 2;
      break;
    default:
      fprintf(stderr, "format not implemented\n");
      return -1;
    }
  avi_out_alloc_bytes = avi_out_bytes;
  if(avi_out_alloc_bytes < 2) avi_out_alloc_bytes = 2;
  nbufs = nbufs_default;
  if(nbufs > 3) nbufs = 3;
  buffers = (void **)(malloc(nbufs*sizeof(void*)));
  for(i=0;i<nbufs;i++)
    buffers[i] = malloc((avifile->width+15)
			*avifile->height*avi_out_alloc_bytes);
  bufcompress=(char *) malloc(6*avifile->width*avifile->height);
  delay = (double)avifile->fps_den / avifile->fps_num;
  t_start = times(&start);
  return 0;
}

static int
grab_close ()
{
  int i;
  xvid_decore(xvid_dec_p.handle,XVID_DEC_DESTROY, NULL, NULL);
  AVI_close(avifile);
  for(i=0;i<nbufs;i++) free(buffers[i]);
  free(buffers);
  if(rgb16tohi240!=NULL) free(rgb16tohi240);
  return 0;
}


static int
grab_overlay (int x, int y, int width, int height, int format,
              struct OVERLAY_CLIP *oc, int count)
{
  return 0;
}


void * avi_extract_frame(int width, int height) {
  struct timeval tv;
  int l;
  XVID_DEC_FRAME dec;

  img++;
  if(img==nbufs) img=0;
  if(width<avifile->width || width>= avifile->width+16
     || height!=avifile->height) {
    fprintf(stderr,
	    "avi_width=%ld requested_width in [%d %d]\n",avifile->width,
	    width-15,width);
    fprintf(stderr,
	    "avi_height=%ld requested_height=%d\n",avifile->height,height);
    fprintf(stderr, "AVI has not the good format\n");
    return NULL;
  }

  if(debug) 
    fprintf(stderr,"extracting frame %ld\n",avifile->video_pos);
  
  if((l=AVI_read_frame(avifile, bufcompress))==-1) {
    AVI_print_error("AVI_read_frame");
    return NULL;
  }

  memset(&dec,0,sizeof(XVID_DEC_FRAME));
  dec.colorspace = avi_out_format;
  dec.image = buffers[img];
  dec.bitstream = bufcompress;
  dec.length = l;
  dec.stride = width;
  
  if(xvid_decore(xvid_dec_p.handle, XVID_DEC_DECODE,&dec,NULL)) {
    fprintf(stderr, "xvid decoding error, the file is probably not XVID\n");
    fprintf(stderr, "  convert with 'mencoder -ovc xvid -oac copy'\n");
    return NULL;
  }
  if(avifile->video_pos>=avifile->video_frames) {
    fprintf(stderr,"END OF AVI FILE REACHED\n");
    AVI_seek_start(avifile);
    if(avibench) {
      int clockpersec = sysconf(_SC_CLK_TCK);
      clock_t t_end;
      struct tms end;
      t_end = times(&end);
      printf("total  %.3lf secs\n",(double)(t_end-t_start)/clockpersec);
      printf("user   %.3lf secs\n",
	     (double)(end.tms_utime-start.tms_utime)/clockpersec);
      printf("system %.3lf secs\n",
	     (double)(end.tms_stime-start.tms_stime)/clockpersec);
      sigint_handler(SIGINT);
    }
  }
  gettimeofday (&tv, NULL);
  videostampmax=tv.tv_usec/1000000.0+tv.tv_sec;
  if(!avibench && videostampmax<videostampmin+delay) {
    usleep(1000000*(delay-videostampmax+videostampmin));
  }
  videostampmin=videostampmax;
  return dec.image;
}  


static void *
grab_scr (void *dest, int width, int height, int single, int dest_xsize)
{
  void *buf;
  
  buf = avi_extract_frame(width,height);
  if(buf == NULL) return NULL;
  if(rgb16tohi240 == NULL)
    fast_memcpy (dest, buf, width * height * avi_out_bytes);
  else {  // RGB08 !!!
    register unsigned char *dest2 = dest;
    register unsigned short * src=buf;
    register int i;
    for(i=width * height-1 ; i>=0; i--) {
      *dest2 = rgb16tohi240[*src];
      dest2++;
      src++;
    }
  }

  return dest;
}

static void *
grab_one (int x11_format, int width, int height)
{
  return NULL;
}

/* ---------------------------------------------------------------------- */

static int
grab_tune (unsigned long freq)
{
  return 0;
}

static int
grab_tuned ()
{
  return 1;
}

static int
grab_input (int input, int norm)
{
  return 0;
}

static int
grab_picture (int color, int bright, int hue, int contrast)
{
  return 0;
}

static int
grab_audio (int mute, int volume, int *mode)
{
  return 0;
}

static int is525(int norm) {
  return norm == 1;
}

static int issecam(int norm) {
  return norm == 2;
}

static void* get_buf(int i){
  return buffers[i];
}

#endif

