/*****************************************************************************
 * hi240torgb_gen.c:
   program which generates hi240torgb.c
   to generate do:
   gcc -o hi240torgb_gen -O2 hi240torgb_gen.c
   ./hi240torgb_gen > hi240torgb.c

   This program capture frames in bgr32 and hi240
   and deduces a probable conversion conversion hi240 -> bgr32.

   There are certainly precise formulas which can compute
   these tables, but anyway I think it is a good example
   of program which manipulates simply a TV-card
 *****************************************************************************
 * $Id: hi240torgb_gen.c,v 1.4 2003/09/02 17:23:06 alainjj Exp $
 *****************************************************************************
 * Copyright (C) 2003 Alainjj
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <math.h>
#include "videodev.h"
#include "videodev2.h"

int bttvfd;

void capture(void *dest, int format) {
  int type,r;
  struct timeval tv;
  fd_set  rdset;
  struct v4l2_format g;
  char *map;
  struct v4l2_requestbuffers reqbufs;
  struct v4l2_buffer gb;
  g.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  g.fmt.pix.width = 768;
  g.fmt.pix.height = 576;
  g.fmt.pix.field = V4L2_FIELD_ANY;
  g.fmt.pix.pixelformat = format;
  if(-1 == ioctl(bttvfd,VIDIOC_S_FMT, &g)) {
    perror("ioctl VIDIOC_S_FMT");
    exit(1);
  }
  reqbufs.count  = 1;
  reqbufs.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  reqbufs.memory = V4L2_MEMORY_MMAP;
  if (-1 == ioctl(bttvfd, VIDIOC_REQBUFS, &reqbufs)) {
    perror("ioctl VIDIOC_REQBUFS");
    exit(1);
  }
  gb.index  = 0;
  gb.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  gb.memory = V4L2_MEMORY_MMAP;
  if (-1 == ioctl(bttvfd, VIDIOC_QUERYBUF, &gb, 0)) {
    perror("ioctl VIDIOC_QUERYBUF");
    exit(1);
  }
  map = mmap(NULL, gb.length, PROT_READ | PROT_WRITE, MAP_SHARED,
	     bttvfd, gb.m.offset);
  if(map == MAP_FAILED) {
    perror("mmap");
    exit(1);
  }
  if(-1 == ioctl(bttvfd,VIDIOC_STREAMON,&type)) {
    perror("ioctl VIDIOC_STREAMON");
    exit(1);
  }
  if (-1 == ioctl (bttvfd, VIDIOC_QBUF,&gb)) {
    perror("ioctl VIDIOC_QBUF");
    exit(1);
  }
  tv.tv_sec  = 5;
  tv.tv_usec = 0;
  FD_ZERO(&rdset);
  FD_SET(bttvfd, &rdset);
  r = select(bttvfd + 1, &rdset, NULL, NULL, &tv);
  if(r==-1) {
    perror("select");
    exit(1);
  }
  if(r==0) {
    fprintf(stderr,"select timeout\n");
    exit(1);
  }
  if (-1 == ioctl (bttvfd, VIDIOC_DQBUF, &gb)){
    perror ("ioctl VIDIOC_DQBUF");
    exit(1);
  }
  memcpy(dest,map,gb.length);
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if(-1 == ioctl(bttvfd,VIDIOC_STREAMOFF,&type)) {
    perror("ioctl VIDIOC_STREAMOFF");
    exit(1);
  }
  munmap(map,gb.length);
}
  
static inline void randcontrol(int x,struct v4l2_queryctrl *ctrl) {
  struct v4l2_control c;
  if(ctrl->id == 0) return;
  c.id = ctrl->id;
  c.value = (float)x/65535*(ctrl->maximum-ctrl->minimum)+ctrl->minimum+0.5;
  if(-1 == ioctl(bttvfd, VIDIOC_S_CTRL, &c)) {
    perror("ioctl VIDIOC_S_CTRL");
    exit(1);
  }
}

struct v4l2_queryctrl queryctrl(int id) {
  struct v4l2_queryctrl	c;
  c.id = id;
  if(-1 == ioctl(bttvfd,VIDIOC_QUERYCTRL,&c)){
    perror("ioctl VIDIOC_QUERYCTRL");
    c.id = 0;
  }
  return c;
}

int main(int nargs, char**argv) {
  unsigned char hi240[768*576];
  int rgb32[768*576];
  int i, j;
  long long n[256];
  double r[256],g[256],b[256];
  double r2[256],g2[256],b2[256];
  int vals = 0;
  struct v4l2_queryctrl	bright,contrast,hue,saturation;
  int nb_frames;
  
  if(nargs!=2) {
    fprintf(stderr, "usage: %s nb_frames\n",argv[0]);
    exit(1);
  }
  
  nb_frames=atoi(argv[1]);
  
  if (-1 == (bttvfd = open ("/dev/video", O_RDWR))) {
    perror("open");
    exit(1);
  }
  bright     = queryctrl(V4L2_CID_BRIGHTNESS);
  contrast   = queryctrl(V4L2_CID_CONTRAST);
  hue        = queryctrl(V4L2_CID_HUE);
  saturation = queryctrl(V4L2_CID_SATURATION);


  for(i=0;i<256;i++) {
    n[i]=0;
    r2[i]=g2[i]=b2[i]=r[i]=g[i]=b[i]=0.0;
  }
  for(j=0;j<nb_frames;j++) {
    randcontrol(rand()&65535,&bright);
    randcontrol(rand()&65535,&contrast);
    randcontrol(rand()&65535,&hue);
    randcontrol(rand()&65535,&saturation);
    capture(rgb32,V4L2_PIX_FMT_BGR32);
    capture(hi240,V4L2_PIX_FMT_HI240);
    for(i=0;i<768*576;i++) {
      int h=hi240[i];
      int k=rgb32[i];
      int bb=k&0x0000FF;
      int gg=(k&0x00FF00)>>8;
      int rr=(k&0xFF0000)>>16;
      n[h]++;
      b[h]+=bb;
      g[h]+=gg;
      r[h]+=rr;  
      b2[h]+=bb*bb;
      g2[h]+=gg*gg;
      r2[h]+=rr*rr;  
    }
    vals=0;
    for(i=0;i<256;i++)
      if(n[i]!=0) vals++;
    fprintf(stderr,"%d frames %d values\r",j,vals);
  }
  printf("/* produced by hi240torgb_gen.c */\n");
  printf("/* (%d frames %5e pixels %d values) */\n",
	 nb_frames,768.0*576*nb_frames,vals);
  printf("unsigned short hi240_to_rgb[256][3]={");
  for(i=0;i<256;i++) {
    int rr,gg,bb;
    //if(i%3==0) printf("\n   ");
    printf("\n   ");
    if(n[i]==0)
      rr=gg=bb=i/255.0*65535+0.5;
    else {
      rr=r[i]/255.0/n[i]*65535+0.5;
      gg=g[i]/255.0/n[i]*65535+0.5;
      bb=b[i]/255.0/n[i]*65535+0.5;
    }
    printf("{%5u, %5u, %5u}",rr,gg,bb);
    if(i!=255) printf(","); else printf(" ");
    printf(" // %3u freq=",i);
    if(n[i]) {
      double freq=n[i]/(768.0*576*nb_frames)*100.0;
      double pr=sqrt((r2[i]-r[i]*r[i]/n[i])/n[i]);
      double pg=sqrt((g2[i]-g[i]*g[i]/n[i])/n[i]);
      double pb=sqrt((b2[i]-b[i]*b[i]/n[i])/n[i]);
      if(n[i]>100000)
	printf("%7.4f%%",freq);
      else
	printf("%8llu",n[i]);
      printf(" prec=%4.1f %4.1f %4.1f",pr,pg,pb);
    } else {
      printf("%8llu",0LL);
    }
  }
  printf("\n};\n");
  exit(0);
}
  

  
