/*****************************************************************************
 * mixer.c: some functions to control a sound mixer
 *****************************************************************************
 * $Id: mixer.c,v 1.15.2.1 2003/11/25 12:20:06 alainjj Exp $
 *****************************************************************************
 * Copyright (C) 2001 Keuleu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************
 *
 * Original code:
 *
 *  (c) 1998 Gerd Knorr <kraxel@cs.tu-berlin.de>
 *
 *****************************************************************************/
#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "config.h"
/* ALSA05 is not supported, with ALSA05 use OSS emulation */
#ifdef HAVE_ALSA
#  if defined(HAVE_ALSA_ASOUNDLIB_H)
#    include <alsa/asoundlib.h>
#  elif defined (HAVE_SYS_ASOUNDLIB_H)
#    include <sys/asoundlib.h>
#  endif
#endif // HAVE_ALSA
#include <linux/soundcard.h>
#include "mixer.h"

extern int debug;
extern int cur_volume;
void set_volume (void);
int have_alsa = 0;

enum {
  LINE=0,
  FRONT,
  REAR,
  IGAIN,
  REARCONTROL,
  MASTER,
  NB_DEVICES
};
	

#ifdef HAVE_ALSA
static char *alsa_group_list[NB_DEVICES] = { (char *) NULL, "PCM", "Wave Surround", "Capture", "SB Live Analog/Digital Output Jack", "Master" };
static snd_mixer_t *handle;
snd_mixer_elem_t *elem[NB_DEVICES] = {NULL, NULL, NULL, NULL, NULL, NULL};
static int oldcap;
#endif
static char *names[] = SOUND_DEVICE_NAMES;
static char *device_list[NB_DEVICES] = { (char *) NULL, "pcm", "dig1", "igain", (char *)NULL, "vol" };
static int mix;
static int dev_id[NB_DEVICES] = { -1, -1, -1, -1, -1, -1 };
static int devmask, recmask, old_recsrc, recsrc;

static int volume[NB_DEVICES];    /* 0x6363 <=> 99 on left channel and 99 on right channel */
static int muted[NB_DEVICES];
static int captured[NB_DEVICES];
static int cur_dev = -1;
static int dev_line = 0, dev_pcm = 0;
static int cur_mute = 0;

#ifdef HAVE_ALSA
static inline long percent_to_alsa(int val, long pmin, long pmax) {
  return pmin + ((pmax - pmin) * val + 50) / 100;  
}
static inline int alsa_to_percent(long val, long pmin, long pmax) {
  long diff = pmax - pmin;
  if(diff == 0) return 50;
  return ((val - pmin) * 100 + (diff / 2)) / diff;
}

int alsa_open(char *device) {
  int j, err;
  if((err = snd_mixer_open( &handle, 0 )) < 0) {
    fprintf(stderr,"snd_mixer_open: %s\n", snd_strerror(err));
    return -1;
  }
  if((err = snd_mixer_attach( handle, "default" )) < 0) {
    fprintf(stderr,"snd_mixer_open: %s\n", snd_strerror(err));
    snd_mixer_close(handle);
    return -1;
  }
  if((err = snd_mixer_selem_register( handle, 0, 0 )) < 0) {
    fprintf(stderr,"snd_mixer_open: %s\n", snd_strerror(err));
    snd_mixer_close(handle);
    return -1;
  }
  if((err = snd_mixer_load( handle )) < 0) {
    fprintf(stderr,"snd_mixer_open: %s\n", snd_strerror(err));
    snd_mixer_close(handle);
    return -1;
  }
  alsa_group_list[0] = (char *) malloc (strlen (device) + 1);
  if (strcasecmp(device, "line")==0)
    strcpy (alsa_group_list[0], "Line");
  else
    strcpy (alsa_group_list[0], device);
  for(j = 0; j < NB_DEVICES; j++) {
    long v,pmin,pmax;
    snd_mixer_selem_id_t *sid;
    if(alsa_group_list[j] == NULL) continue;
    snd_mixer_selem_id_alloca(&sid);
    snd_mixer_selem_id_set_index(sid,0);
    snd_mixer_selem_id_set_name(sid, alsa_group_list[j]);
    elem[j] = snd_mixer_find_selem(handle, sid);
    if(elem[j] == NULL) continue;
    if(j == IGAIN) {
      snd_mixer_selem_get_capture_volume_range(elem[j],&pmin,&pmax);
      snd_mixer_selem_get_capture_volume(elem[j], 0, &v);
    } else {
      snd_mixer_selem_set_playback_switch_all(elem[j], 1);
      snd_mixer_selem_get_playback_volume_range(elem[j],&pmin,&pmax);
      snd_mixer_selem_get_playback_volume(elem[j], 0, &v);
    }
    volume[j] = alsa_to_percent(v, pmin, pmax);
    if(volume[j] == 0) {
      volume[j] = (j == IGAIN) ? 10 : 68;
      v = percent_to_alsa(volume[j], pmin, pmax);
      if(j == IGAIN)
	snd_mixer_selem_set_capture_volume_all(elem[j], v);
      else
	snd_mixer_selem_set_playback_volume_all(elem[j], v);
    }
    if(debug)
      fprintf(stderr,"ALSAMIXER: %-8s=%3ld (%3d%%) min=%3ld max=%3ld\n",
	      alsa_group_list[j], v, volume[j], pmin, pmax);
  }
  cur_dev = 0;
  
  if (elem[1] != NULL)
    dev_pcm = 1;
  else if(elem[2] != NULL)
    dev_pcm = 2;
  snd_mixer_selem_get_capture_switch(elem[dev_line],0,&oldcap);
  return (NULL != elem[0]) ? 0 : -1;
}

#endif




int
mixer_open (char *filename, char *device)
{
  int i, j, dix = 0x0a0a;
  
#ifdef HAVE_ALSA
  if(have_alsa != -1)
    if(alsa_open(device) != -1) {
      have_alsa = 1;
      return 0;
    }
#endif
  
  have_alsa = 0;
  device_list[0] = (char *) malloc (strlen (device) + 1);
  strcpy (device_list[0], device);

  if (-1 == (mix = open (filename, O_RDONLY)))
    {
      perror ("mixer open");
      return -1;
    }

  if (-1 == ioctl (mix, MIXER_READ (SOUND_MIXER_DEVMASK), &devmask))
    {
      perror ("mixer read devmask");
      return -1;
    }
  //printf("devmask = 0x%04X\n", devmask);    

  if (-1 == ioctl (mix, MIXER_READ (SOUND_MIXER_RECMASK), &recmask))
    {
      perror ("mixer read recmask");
      return -1;
    }
  //printf("recmask = 0x%04X\n", recmask);    

  for (j = 0; j < NB_DEVICES; j++)
    {
      if (device_list[j] == (char *)NULL)
        continue;

      for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
        {
          if (debug)
            printf("mixer names [%d] = '%s'\n", i, names[i]);

          if ((1 << i) & devmask
              && strcasecmp (names[i], device_list[j]) == 0)
            {
              if (-1 == ioctl (mix, MIXER_READ (i), &volume[j]))
                {
                  perror ("mixer read volume");
                  return -1;
                }
              else
                {
                  dev_id[j] = i;
                  muted[j] = 0;
                  if (volume[j] == 0)
                    {
                      if (j == IGAIN)
                        volume[j] = dix;
                      else
                        volume[j] = 0x4444;
                      if (-1 == ioctl (mix, MIXER_WRITE (i), &volume[j]))
                        {
                          perror ("mixer write volume");
                          return -1;
                        }
                    }

                  if (debug)
                    printf
                      ("mixer: %s (%d) opened with volume = %d = 0x%04X\n",
                       device_list[j], dev_id[j], volume[j], volume[j]);
                }
              break;
            }
        }
    }



  if (-1 == dev_id[0])
    {
      fprintf (stderr, "mixer: hav'nt found device '%s'\nmixer: available: ",
               device);
      for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
        if ((1 << i) & devmask)
          fprintf (stderr, " '%s'", names[i]);
      fprintf (stderr, "\n");
    }

  if (-1 == ioctl (mix, MIXER_READ (SOUND_MIXER_RECSRC), &recsrc))
    {
      perror ("mixer read recsrc");
      return -1;
    }
  old_recsrc = recsrc;
  for (i = 0; i < NB_DEVICES; i++)
    {
      captured[i] = 0;
      if ((1 << dev_id[i]) & recmask)
        {
          if ((1 << dev_id[i]) & recsrc)
            captured[i] = 1;
        }
      if (debug)
        printf ("mixer: mixer_open: captured (%s) = %d\n", device_list[i],
                captured[i]);
    }

  // Make sure IGAIN isn't muted
  if (-1 == ioctl (mix, MIXER_READ (dev_id[IGAIN]), &volume[IGAIN]))
    fprintf(stderr, "mixer: mixer_open: couldn't read %s volume\n", device_list[IGAIN]);

  if(debug)
    fprintf(stderr, "****************igain volume = 0x%04X\n", volume[IGAIN]);

  if ( ((volume[IGAIN] & 0xFF) == 0) || ((volume[IGAIN] & 0xFF00)>>8 == 0) )
    {
      if (-1 == ioctl (mix, MIXER_WRITE (dev_id[cur_dev]), &dix))
        fprintf(stderr, "mixer: mixer_open: couldn't write %s volume\n", device_list[IGAIN]);
  if (debug)
    fprintf(stderr, "mixer: mixer_open: OSS setting IGAIN volume to %04X\n", dix);
    }

  cur_dev = 0;

  if (dev_id[1] != -1)
    dev_pcm = 1;
  else if(dev_id[2] != -1)
    dev_pcm = 2;

  return (-1 != dev_id[0]) ? 0 : -1;
}

void
mixer_close (void)
{
  int i;

  if (debug)
    fprintf (stderr, "mixer: close\n");

  mixer_set_mute(0);

#ifdef HAVE_ALSA
  if(have_alsa) {
    snd_mixer_close(handle);
    return;
  }
#endif

  close (mix);

  for (i = 0; i < NB_DEVICES; i++)
    dev_id[i] = -1;
  cur_dev = -1;

  if (device_list[0])
    {
      free (device_list[0]);
      device_list[0] = (char *) NULL;
    }
}

int
mixer_get_volume (void)
{
  //printf("mixer: mixer_get_volume: volume (%s) = 0x%02X\n", device_list[cur_dev], volume[cur_dev]&0x7f);
#ifdef HAVE_ALSA
  if(have_alsa)
    return (NULL == elem[cur_dev]) ? -1 : (volume[cur_dev]);
  else
#endif
    return (-1 == dev_id[cur_dev]) ? -1 : (volume[cur_dev] & 0x7f);
}

int
mixer_set_volume (int val)
{
#ifdef HAVE_ALSA
  if(have_alsa)
    {
      long v, pmin, pmax;
      if (NULL == elem[cur_dev])
	return -1;
      snd_mixer_selem_get_playback_volume_range(elem[cur_dev],&pmin,&pmax);
      v = percent_to_alsa(val, pmin, pmax);
      if(debug)
	fprintf(stderr,"cur_dev=%d %ld %ld %ld\n",cur_dev,v,pmin,pmax);
      snd_mixer_selem_set_playback_volume_all(elem[cur_dev], v);
      volume[cur_dev] = val;
    }
  else
#endif
    {
      if (-1 == dev_id[cur_dev])
	return -1;
      
      //printf("mixer: mixer_set_volume: val = %d = 0x%02X\n", val, val);
      val &= 0x7f;
      volume[cur_dev] = val | (val << 8);;
      //printf("mixer: mixer_set_volume: volume (%s) = %d = 0x%04X\n", device_list[cur_dev],volume[cur_dev], volume[cur_dev]);
      
      if (-1 == ioctl (mix, MIXER_WRITE (dev_id[cur_dev]), &volume[cur_dev]))
	{
	  perror ("mixer write volume");
	  return -1;
	}
    }
  
  muted[cur_dev] = 0;
  return 0;
}


int mixer_set_mute (int val)
{
  int r;
  if(val)
    r = mixer_mute();
  else
    r = mixer_unmute();
  if(r == -1) return -1;
  cur_mute = val;
  return 0;
}


int
mixer_mute (void)
{
  int zero = 0;
  
#ifdef HAVE_ALSA
  if(have_alsa)
    {
      if(NULL == elem[cur_dev]) return -1;
      snd_mixer_selem_set_playback_switch_all(elem[cur_dev], 0);
    }
  else
#endif
    {
      if (-1 == dev_id[cur_dev])
	return -1;
      
      if (-1 == ioctl (mix, MIXER_WRITE (dev_id[cur_dev]), &zero))
	return -1;
    }

  muted[cur_dev] = 1;

  return 0;
}

int
mixer_unmute (void)
{

#ifdef HAVE_ALSA
  if(have_alsa)
    {
      if(NULL == elem[cur_dev]) return -1;
      snd_mixer_selem_set_playback_switch_all(elem[cur_dev], 1);
    }
  else
#endif
    {
      if (-1 == dev_id[cur_dev])
	return -1;
      
      if (-1 == ioctl (mix, MIXER_WRITE (dev_id[cur_dev]), &volume[cur_dev]))
	return -1;
    }

  muted[cur_dev] = 0;

  return 0;
}

int
mixer_get_muted (void)
{
#ifdef HAVE_ALSA
  if(have_alsa)
    return (NULL == elem[cur_dev]) ? -1 : muted[cur_dev];
  else
#endif
    return (-1 == dev_id[cur_dev]) ? -1 : muted[cur_dev];
}

int
mixer_capture (void)
{
#ifdef HAVE_ALSA
  if(have_alsa)
    {
      if(NULL == elem[cur_dev]) return -1;
      snd_mixer_selem_set_capture_switch_all(elem[dev_line], 1);
      cur_volume = volume[dev_pcm];
    }
  else
#endif
    {
      if (-1 == dev_id[dev_pcm])
	return -1;
      
      if (!((1 << dev_id[dev_line]) & recmask))
	return -1;
      
      recsrc = (1 << dev_id[dev_line]);
      if (-1 == ioctl (mix, MIXER_WRITE (SOUND_MIXER_RECSRC), &recsrc))
	return -1;
      cur_volume = volume[dev_pcm]&0x7F;
    }

  captured[dev_line] = 1;

  cur_dev = dev_pcm;
  
  set_volume();

  return 0;
}


int
mixer_uncapture (void)
{
#ifdef HAVE_ALSA
  if(have_alsa)
    {
      if(NULL == elem[cur_dev]) return -1;
      snd_mixer_selem_set_capture_switch_all(elem[dev_line], oldcap);
      cur_volume = volume[dev_line];
    }
  else
#endif
  {
    if (-1 == dev_id[dev_line])
      return -1;
    
    recsrc = old_recsrc;
    if (-1 == ioctl (mix, MIXER_WRITE (SOUND_MIXER_RECSRC), &recsrc))
      return -1;
    cur_volume = volume[dev_line]&0x7F;
  }

  captured[dev_line] = 0;
  
  cur_dev = dev_line;

  set_volume();

  return 0;
}

int
mixer_get_captured (void)
{
#ifdef HAVE_ALSA
  if(have_alsa)
    return (NULL == elem[cur_dev]) ? -1 : captured[cur_dev];
  else
#endif
    return (-1 == dev_id[cur_dev]) ? -1 : captured[cur_dev];
}

int
mixer_toggle_device (void)
{
  cur_dev++;
  cur_dev %= NB_DEVICES;
  return 0;
}

